////////////////////////////////////////////////////////////////////////////
//
// FanPWMvs].
//
// Copyright by rubidesu 2022.
// http://www.gatelink.co.jp/hw/
//
// 2022.05.05 V1.00.
//
////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <conio.h>
#include <math.h>
#include <windows.h>
#include <process.h>

HANDLE hCom = NULL;
OVERLAPPED ovl;

BOOL boThreadExit = FALSE;

#define BUFFER_SIZE	128
BYTE bRcvBuf[BUFFER_SIZE];	//Mobt@.
BYTE bRcvPnt = 0;		//Mobt@̃|C^.

BYTE bFanDuty = 0;
DWORD dwFreq = 0;

int iWaitSec = 0;
int iWaitSecCnt = 0;

#define STEP_MAX	21
int iStepNum = 0;

////////////////////////////////////////////////////////////////////////////
BOOL ComPortOpen(unsigned int port, unsigned int baud_rate)
{
	char cPort[32];

	if(port == 0){
		return FALSE;
	}

	sprintf(cPort, "\\\\.\\COM%d", port);
	hCom = ::CreateFile(cPort,
						GENERIC_WRITE|GENERIC_READ,
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						NULL,
						OPEN_EXISTING,
						FILE_FLAG_OVERLAPPED,
						NULL);

	if(INVALID_HANDLE_VALUE==hCom)
	{
		hCom = NULL;
		return FALSE;
	}

	DCB dcb;
	::GetCommState(hCom,&dcb);
	dcb.BaudRate = baud_rate;
	dcb.ByteSize = 8;
	dcb.Parity = 0;
	dcb.fParity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	dcb.fBinary = TRUE;
	::SetCommState(hCom,&dcb);

	COMMTIMEOUTS cto;
	cto.ReadIntervalTimeout = 100;
	cto.ReadTotalTimeoutConstant = 10;
	cto.ReadTotalTimeoutMultiplier = 10;
	cto.WriteTotalTimeoutConstant = 10;
	cto.WriteTotalTimeoutMultiplier = 10;
	::SetCommTimeouts(hCom,&cto);

	return TRUE;
}

////////////////////////////////////////////////////////////////////////////
void ComPortClose()
{
	if(hCom == NULL){
		return;
	}

	::CloseHandle(hCom);

	hCom = NULL;
}

////////////////////////////////////////////////////////////////////////////
void ComPortSendCommand(unsigned char data1)
{
	DWORD bytes = 0;
	unsigned char send_char[1];

	DWORD dwRcvNum;
	DWORD dwErr;
	DWORD dwEndtime;

	send_char[0] = data1;

	if(WriteFile(hCom, send_char, 1, &bytes, &ovl) == 0){
		dwErr = GetLastError();
		dwRcvNum = 0;
		if(dwErr == ERROR_IO_PENDING){
			dwEndtime = GetTickCount()+3000;
			while(!GetOverlappedResult(hCom, &ovl, &dwRcvNum, TRUE)){
				if(GetTickCount() > dwEndtime){
					break;
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////
void PrintData(void)
{
	printf("%d\t%d\n",
		bFanDuty * 5,
		dwFreq * 60 / 2
		);
}

////////////////////////////////////////////////////////////////////////////
void AnalyzeData(void)
{
	BYTE bTemp[64];

	memcpy((void*)&bTemp, (void*)&bRcvBuf, 16);

	bFanDuty = bTemp[2];

	dwFreq  = bTemp[ 5] * 1000000000;
	dwFreq += bTemp[ 6] * 100000000;
	dwFreq += bTemp[ 7] * 10000000;
	dwFreq += bTemp[ 8] * 1000000;
	dwFreq += bTemp[ 9] * 100000;
	dwFreq += bTemp[10] * 10000;
	dwFreq += bTemp[11] * 1000;
	dwFreq += bTemp[12] * 100;
	dwFreq += bTemp[13] * 10;
	dwFreq += bTemp[14] * 1;
}

////////////////////////////////////////////////////////////////////////////
VOID StackData(BYTE bRcv)
{
	if(bRcvPnt > (BUFFER_SIZE - 1)){
		//obt@I[ot[.
		bRcvPnt = 0;
		bRcvBuf[bRcvPnt] = bRcv;
	}
	else if(bRcv == 0xfe){
		//s̎M.

		bRcvBuf[bRcvPnt] = bRcv;
		bRcvPnt = bRcvPnt + 1;

		if((bRcvBuf[0] == 0xfd) && (bRcvBuf[15] == 0xfe)){
			fprintf(stderr, "*");

			iWaitSecCnt++;

			if(iWaitSecCnt >= iWaitSec){
				AnalyzeData();

				fprintf(stderr, " %drpm\n", dwFreq * 60 / 2);

				PrintData();

				if(iStepNum < STEP_MAX){
					fprintf(stderr, "Set Duty %3d%% ", iStepNum * 5);
					ComPortSendCommand(0x80 + iStepNum);
				}
				iStepNum++;

				iWaitSecCnt = 0;
			}
		}
		else{
		}

		bRcvPnt = 0;
	}
	else{
		//ꕶ̎M.
		bRcvBuf[bRcvPnt] = bRcv;
		bRcvPnt = bRcvPnt + 1;
	}
}

////////////////////////////////////////////////////////////////////////////
unsigned __stdcall ReceiveThread(void *pArg)
{
	DWORD event;
	BYTE buffer[32];
	DWORD dwRcvNum;
	DWORD dwErr;
	DWORD dwEndtime;

	memset(&ovl, 0, sizeof(OVERLAPPED));
	ovl.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	SetCommMask(hCom, EV_RXCHAR);

	for(;;){
		WaitCommEvent(hCom, &event, NULL);

		if(boThreadExit == TRUE){
			break;
		}

		if((event & EV_RXCHAR) == EV_RXCHAR){
			dwErr = -1;
			dwRcvNum = 0;

			if(ReadFile(hCom, buffer, 32, &dwRcvNum, &ovl) == 0){
				dwErr = GetLastError();
				dwRcvNum = 0;
				if(dwErr == ERROR_IO_PENDING){
					dwEndtime = GetTickCount()+3000;
					while(!GetOverlappedResult(hCom, &ovl, &dwRcvNum, TRUE)){
						if(GetTickCount() > dwEndtime){
							break;
						}
					}
				}
			}
			if(dwErr==0 || dwErr==ERROR_IO_PENDING){
				if(dwRcvNum != 0){
					for(int i = 0; i < dwRcvNum; i++){
						StackData(buffer[i]);
					}
				}
			}

			ResetEvent(ovl.hEvent);
		}
	}

	boThreadExit = FALSE;

	_endthreadex(0);
	return 0;
}

////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
	BYTE cKey;
	int iComPort;
	HANDLE hThread;

	if(argc < 3){
		fprintf(stderr, "Usage : PWMSweep ComPort WaitSec\r\n");
		return 0;
	}

	iComPort = atoi(argv[1]);
	iWaitSec = atoi(argv[2]);

	if(iWaitSec < 1){
		fprintf(stderr, "Usage : PWMSweep ComPort WaitSec\r\n");
		fprintf(stderr, "WaitSec must be 1 or greater.\r\n");
		return 0;
	}

	if(ComPortOpen(iComPort, 9600) == FALSE){
		fprintf(stderr, "Can't Open COM%d\r\n", iComPort);
		return 0;
	}

	if((hThread = (HANDLE)_beginthreadex(NULL, 0, &ReceiveThread, NULL, 0, NULL)) == 0){
		fprintf(stderr, "Can't Run Thread.\r\n");
		ComPortClose();
		return 0;
	}

	fprintf(stderr, "ComPort=%d, WaitSec=%d, 'Q':Exit.\r\n", iComPort, iWaitSec);
	printf("Duty(%%)\tSpeed(rpm)\n");

	ComPortSendCommand(0x02);	//Set 4Pin Mode.

	fprintf(stderr, "Set Duty %3d%% ", iStepNum * 5);
	ComPortSendCommand(0x80 + iStepNum);	//Duty 0%.
	iStepNum++;

	for(;;){
		if(kbhit() != 0){
			cKey = getch();
			if((cKey == 'q') || (cKey == 'Q')){
				goto ExitProc;
			}
		}

		if(iStepNum > STEP_MAX){
			break;
		}

		Sleep(250);
	}

ExitProc:

	ComPortSendCommand(0x80 + 0);	//Duty 0%.

	boThreadExit = TRUE;
	SetCommMask(hCom, EV_RXCHAR);

	WaitForSingleObject(hThread, INFINITE);
	CloseHandle(hThread);

	ComPortClose();

	return 0;
}
